/*
  thorfinn@binf.ku.dk april 15 2012
  class to call genotypes
 */

#include "shared.h"
#include <cmath>
#include "analysisFunction.h"

//filename of dumped file
#define GENO ".geno"

//stuff related to genotypecalling
#define GENO_MAJOR_MINOR 0x1 // write the major and the minor
#define GENO_PRINT 0x2 //print the called genotype 0,1,2
#define GENO_ALLELE 0x4 //print the called genotype AA, AC AG ...
#define GENO_ALL_POST 0x8 //print the all posterior
#define GENO_WRITE_POST 16 //write the post of the called genotype
#define GENO_FOR_COVAR 32 //binary dump of the posteriors.

typedef struct{
  int **dat;
}genoCalls;


class callGenotypes:public general{
public:
  int doGeno;
  int doPost;
  int doMaf;
  float postCutoff;
  //none optional stuff
  FILE *outfile;
  callGenotypes(const char *outfiles,argStruct *arguments,int inputtype);
  ~callGenotypes();
  void getOptions(argStruct *arguments);
  void run(funkyPars  *pars);
  void clean(funkyPars *pars);
  void print(funkyPars *pars);
  void printArg(FILE *argFile);
  void getGeno(funkyPars *pars);
  void printGeno(funkyPars *pars);
  
};
void callGenotypes::printArg(FILE *argFile){
  fprintf(argFile,"%s:\n\n",__FILE__);
  fprintf(argFile,"-doGeno\t%d\n",doGeno);
  fprintf(argFile,"\t->1: Thorfinn add something\n");
  fprintf(argFile,"\t->2: Thorfinn add something\n");
  fprintf(argFile,"\n");

}


void callGenotypes::getOptions(argStruct *arguments){
  //default
  doMaf=0;
  doGeno=0;
  doPost=0;
  postCutoff=-1;
  //from command line
  doGeno=angsd::getArg("-doGeno",doGeno,arguments);
 if(doGeno==-999){
    doGeno=0;
    printArg(stdout);
    exit(0);
   }
 
  if(doGeno==0)
    return;
  doMaf=angsd::getArg("-doMaf",doMaf,arguments);
  doPost=angsd::getArg("-doPost",doPost,arguments);
  postCutoff=angsd::getArg("-postCutoff",postCutoff,arguments);
  printArg(arguments->argumentFile);

}

callGenotypes::callGenotypes(const char *outfiles,argStruct *arguments,int inputtype){

  getOptions(arguments);

  if(doGeno==0)
    return;


  //inputtype 0) soap 1) samglf 2) samglfClean 3) tglf 4)sim type 5) beagle
  /*
  if((doGeno&GENO_MAF_PRIOR ) &&(doMaf==0)){
    fprintf(stderr,"\n\t-> You need a maf estimator for this genotypecaller\n\t\n\n");
    //    args->doMaf = args->doMaf|MAF_KNOWN_EM;
    fflush(stderr);
    exit(0);
  }
  */

  if(arguments->inputtype!=5&doPost==0){
    fprintf(stderr,"\n\t-> You need -doPost to call genotypes \n");
    exit(0);

  }
  fprintf(stderr,"Calling genotypes: %s\n",outfiles);
  //make output files
  const char* postfix;
  postfix=GENO;
  outfile = openFile(outfiles,postfix);
  FILE *oFileGenos=outfile;

  
}


callGenotypes::~callGenotypes(){
  //  fprintf(stderr,"[%s] calling destructor\n",__FUNCTION__);
  //  fflush(stderr);
  if(doGeno==0)
    return;

  fclose(outfile);
}


void callGenotypes::getGeno(funkyPars *pars){
  //pp is eiter the genotype likelihoods or the posterior probablity
  genoCalls *geno =(genoCalls *) pars->extras[index];
  for(int s=0;s<pars->numSites;s++){
    if(pars->keepSites==0)
      continue;
    for( int i =0;i<pars->nInd;i++){

      double maxPP=pars->post[s][i*3+0];
      int maxGeno=0;
      if(pars->post[s][i*3+1]>maxPP){
	maxPP=pars->post[s][i*3+1];
	maxGeno=1;
      }
      if(pars->post[s][i*3+2]>maxPP){
	maxPP=pars->post[s][i*3+2];
	maxGeno=2;
      }
      if(maxPP<postCutoff)
	maxGeno=-1;
      geno->dat[s][i]=maxGeno;
    }
  }
}



void callGenotypes::printGeno(funkyPars *pars){

  genoCalls *geno =(genoCalls *) pars->extras[index];

  if(pars->keepSites==NULL){
    fprintf(stderr,"\t->[%s] We would expect to have an array of keepsites\n",__FUNCTION__);
    exit(0);
  }

  for(int s=0;s<pars->numSites;s++) {
    if(pars->keepSites[s]==0)
      continue;
    if(!(doGeno&GENO_FOR_COVAR))
      fprintf(outfile,"%s\t%d\t",pars->sites[s].chromo,pars->sites[s].position+1);
    if(doGeno&GENO_MAJOR_MINOR&&!(doGeno&GENO_FOR_COVAR))
      fprintf(outfile,"%c\t%c\t",intToRef[pars->major[s]],intToRef[pars->minor[s]]);
    if(doGeno&GENO_FOR_COVAR){
      fwrite(pars->post[s],sizeof(double),3*pars->nInd,outfile);
      continue;
    }

    for(int i=0;i<pars->nInd;i++){
      if(doGeno&GENO_PRINT)
	fprintf(outfile,"%d\t",geno->dat[s][i]);
      if(doGeno&GENO_ALLELE){
	if(geno->dat[s][i]==0)
	  fprintf(outfile,"%c%c\t",intToRef[pars->major[s]],intToRef[pars->major[s]]);
	else if(geno->dat[s][i]==1)
	  fprintf(outfile,"%c%c\t",intToRef[pars->major[s]],intToRef[pars->minor[s]]);
	else if(geno->dat[s][i]==2)
	  fprintf(outfile,"%c%c\t",intToRef[pars->minor[s]],intToRef[pars->minor[s]]);
	else if(geno->dat[s][i]==-1)
	  fprintf(outfile,"NN\t");

      }
      if(doGeno&GENO_WRITE_POST)
	fprintf(outfile,"%f\t",pars->post[s][i*3+geno->dat[s][i]]);
      if(doGeno&GENO_ALL_POST)
	for(int n=0;n<3;n++)
	  fprintf(outfile,"%f\t",pars->post[s][i*3+n]);
    }
    fprintf(outfile,"\n");

  }

}




void callGenotypes::clean(funkyPars *pars){
  if(doGeno==0)
    return;
  genoCalls *geno =(genoCalls *) pars->extras[index];


  for(int s=0;s<pars->numSites;s++)
    delete[] geno->dat[s];
  delete[] geno->dat;

  delete geno;
}


void callGenotypes::print(funkyPars *pars){
  if(doGeno==0)
    return;
  
  genoCalls *geno =(genoCalls *) pars->extras[index];

  printGeno(pars);
  
}

//do nothing every thing is precalucated 
void callGenotypes::run(funkyPars *pars){

  if(doGeno==0)
    return;
  //allocate genoCall struct
  genoCalls *geno = new genoCalls;
  geno->dat = new int*[pars->numSites];   
  for(int s=0;s<pars->numSites;s++)
    geno->dat[s]=new int[pars->nInd];
  //genoCall struct to pars
  pars->extras[index] = geno;


  getGeno(pars);
}

